package de.zalando.aruha.nakadi.service.subscription.state;

import de.zalando.aruha.nakadi.service.subscription.KafkaClient;
import de.zalando.aruha.nakadi.service.subscription.StreamParameters;
import de.zalando.aruha.nakadi.service.subscription.StreamingContext;
import de.zalando.aruha.nakadi.service.subscription.model.Partition;
import de.zalando.aruha.nakadi.service.subscription.zk.ZKSubscription;
import de.zalando.aruha.nakadi.service.subscription.zk.ZkSubscriptionClient;
import java.util.Collections;
import org.apache.kafka.clients.consumer.Consumer;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

public class StreamingStateTest {

    private StreamingState state;
    private KafkaClient kafkaMock;
    private ZkSubscriptionClient zkMock;

    private static final String SESSION_ID = "ssid";

    @Before
    public void prepareMocks() {
        state = new StreamingState();

        final StreamingContext contextMock = Mockito.mock(StreamingContext.class);

        Mockito.when(contextMock.getSessionId()).thenReturn(SESSION_ID);
        Mockito.when(contextMock.isInState(Mockito.same(state))).thenReturn(true);

        kafkaMock = Mockito.mock(KafkaClient.class);
        Mockito.when(contextMock.getKafkaClient()).thenReturn(kafkaMock);

        zkMock = Mockito.mock(ZkSubscriptionClient.class);
        Mockito.when(contextMock.getZkClient()).thenReturn(zkMock);

        final StreamParameters spMock = Mockito.mock(StreamParameters.class);
        Mockito.when(contextMock.getParameters()).thenReturn(spMock);

        state.setContext(contextMock, "test");
    }

    @Test
    public void ensureTopologyEventListenerRegisteredRefreshedClosed() {
        final ZKSubscription topologySubscription = Mockito.mock(ZKSubscription.class);
        Mockito.when(zkMock.subscribeForTopologyChanges(Mockito.anyObject())).thenReturn(topologySubscription);

        state.onEnter();

        Mockito.verify(zkMock, Mockito.times(1)).subscribeForTopologyChanges(Mockito.any());
        Mockito.verify(topologySubscription, Mockito.times(0)).refresh();

        state.topologyChanged();

        Mockito.verify(topologySubscription, Mockito.times(1)).refresh();
        Mockito.verify(topologySubscription, Mockito.times(0)).cancel();

        state.onExit();

        Mockito.verify(topologySubscription, Mockito.times(1)).cancel();
        // verify that no new locks created.
        Mockito.verify(zkMock, Mockito.times(1)).subscribeForTopologyChanges(Mockito.any());
    }

    @Test
    public void ensureOffsetsSubscriptionsAreRefreshedAndClosed() {
        final ZKSubscription offsetSubscription = Mockito.mock(ZKSubscription.class);

        final Partition.PartitionKey pk = new Partition.PartitionKey("t", "0");
        Mockito.when(zkMock.subscribeForOffsetChanges(Mockito.eq(pk), Mockito.any())).thenReturn(offsetSubscription);

        final Consumer kafkaConsumer = Mockito.mock(Consumer.class);
        Mockito.when(kafkaConsumer.assignment()).thenReturn(Collections.emptySet());
        Mockito.when(kafkaMock.createKafkaConsumer()).thenReturn(kafkaConsumer);

        state.onEnter();

        state.refreshTopologyUnlocked(new Partition[]{new Partition(pk, SESSION_ID, null, Partition.State.ASSIGNED)});

        Mockito.verify(zkMock, Mockito.times(1)).subscribeForOffsetChanges(Mockito.eq(pk), Mockito.any());
        Mockito.verify(offsetSubscription, Mockito.times(0)).cancel();
        Mockito.verify(offsetSubscription, Mockito.times(0)).refresh();

        state.offsetChanged(pk);
        Mockito.verify(zkMock, Mockito.times(1)).subscribeForOffsetChanges(Mockito.eq(pk), Mockito.any());
        Mockito.verify(offsetSubscription, Mockito.times(0)).cancel();
        Mockito.verify(offsetSubscription, Mockito.times(1)).refresh();

        // Verify that offset change listener is removed
        state.refreshTopologyUnlocked(new Partition[0]);
        Mockito.verify(zkMock, Mockito.times(1)).subscribeForOffsetChanges(Mockito.eq(pk), Mockito.any());
        Mockito.verify(offsetSubscription, Mockito.times(1)).cancel();
        Mockito.verify(offsetSubscription, Mockito.times(1)).refresh();
    }

}